# Generated by Django 2.2.19 on 2021-03-15 10:35

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RSAKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('priv_key', models.TextField(help_text='Paste your private RSA Key here.', unique=True, verbose_name='Key')),
            ],
            options={
                'verbose_name': 'RSA Key',
                'verbose_name_plural': 'RSA Keys',
            },
        ),
    ]
