# Generated by Django 4.2.3 on 2023-10-17 19:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0017_alter_activity_urlid_alter_follower_urlid_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='payload',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='activity',
            name='response_body',
            field=models.TextField(null=True),
        ),
    ]
